#include "adam.h"
#include <algorithm>
#include <math.h>
#include <iostream>
#include <omp.h>
#include "adept.h"
#include "adept_arrays.h"

double sum(double *s1,unsigned num){
    double s = 0;
    for(unsigned i = 0; i < num;i++){
        s += s1[i];
    }
    return s;
}

#define max(a,b) (a>b?a:b)

adam::adam(Genetic *pp,nikolas::rnn **p,Data initialx,int max,Dataset *tr,unsigned maxTh)
 : threads(maxTh)
{
	pgen=pp;
    d = tr;
    pf = p;
    //pf->setParameters(initialx);
    maxSteps = unsigned(max);
    grad = new double[initialx.size()];
    m = new double[initialx.size()];
    v = new double[initialx.size()];
    for(unsigned i = 0; i < initialx.size();i++){
        m[i]=0;
        v[i]=0;
    }
    newW = initialx;
    privateRnn = new nikolas::rnn(pf[0]->getLexicSize(),pf[0]->getHiddenSize());
    adeptGrad = new double[initialx.size()];
}

adam::~adam(){
    delete [] grad;
    delete [] m;
    delete [] v;
    delete [] adeptGrad;
    delete privateRnn;
}

void adam::setAdeptGrad(){
	
    using adept::adouble;
    std::vector<adouble> gr(newW.size());
    //gr.resize(newW.size());
    for(unsigned i = 0; i < newW.size();i++) gr[i] = newW[i];
    stack.new_recording();
    adouble y = comp(gr);
    y.set_gradient(1.0);
    stack.compute_adjoint();
    //stack.compute_tangent_linear();
    for(unsigned i = 0; i < gr.size();i++){
        adeptGrad[i] = gr[i].get_gradient() ;
    }
    stack.clear_gradients();
    //std::cout << std::endl;
}

void adam::run(){
    double att,fit,diff;
    int dropout = 0;
    for(unsigned i = 1; i < maxSteps; i++){
        //dropout = 0;
        setAdeptGrad();
#ifdef ADAMAX
        float sum = 0;
        att = a * sqrt( 1.0 - pow(b2,double(i)) ) / (1.0 - pow(b1,double(i)) );
        for(unsigned j = 0; j < newW.size();j++){
            m[j] = m[j]*b1 + (1.0-b1) * adeptGrad[i];
            v[j] = max(v[j]*b2,abs(adeptGrad[i]));
            newW[j] -= (a / (1.0 - pow(b1,double(i))) )* m[j] / (v[j] + 1e-8);
            sum+=adeptGrad[j]*adeptGrad[j];
        }
#endif
#ifdef ADAM
        float sum = 0;
        att = a * sqrt( 1.0 - pow(b2,double(i)) ) / (1.0 - pow(b1,double(i)) );
        for(unsigned j = 0; j < newW.size();j++){
            m[j] = m[j]*b1 + (1.0-b1) * adeptGrad[j];
            v[j] = v[j]*b2 + (1.0-b2) * adeptGrad[j]*adeptGrad[j];
            diff = att * m[j] / ( sqrt(v[j]) + 1e-8);
            if(!isnan(diff)&&(diff!=0.0))
                newW[j] -= diff;
            else{
                dropout++;
                newW[j]=0;
            }
            sum+=adeptGrad[j]*adeptGrad[j];
        }//*/
#endif
        if(i%(maxSteps/5)==0 && Vars.verbose){
            privateRnn->setParameters(newW);
            fit = getFit();
#pragma omp critical
            std::cout << "Iteration " << i << " : Fitness estimation " << fit << " Gradient norm "<< sqrt(sum) << " THREAD: " << omp_get_thread_num()<< std::endl;
        }
        //
        //*
        offset+=MINIBATCH_SIZE;
        if(offset>1.0-MINIBATCH_SIZE) {
            offset=0;
        }//*/
    }
    std::cout << "Iteration " << maxSteps << " : Fitness estimation " << fit
              << std::endl;
}


double adam::getFit(){
    double retVal = 0.0;
    int total = 0;
    for(unsigned i = offset*double(d->positiveLines(TrainData)); i < (MINIBATCH_SIZE+offset) * double(d->positiveLines(TrainData));i++){
        retVal += privateRnn->computeFit(d->getPositiveLine(i,TrainData));
        total += d->getPositiveLine(i,TrainData).size()-1;
    }
    for(unsigned i = offset*double(d->negativeLines()); i < (MINIBATCH_SIZE+offset) * double(d->negativeLines());i++){
        retVal += (d->getNegativeLine(i,TrainData).size() - privateRnn->computeFit(d->getNegativeLine(i,TrainData)));
        total += d->getNegativeLine(i,TrainData).size()-1;
    }
    //std::cout << retVal/total << " " << retVal << "/" << total << " ";
    return  retVal/(1.0*total);
}

adouble adam::comp(vector<adept::adouble> x){
    adouble retVal = 0.0;
    adouble total = 0;
    //pf[omp_get_thread_num()]->setParameters(x);
    privateRnn->setParameters(x);
    for(int i = int( offset*double(d->positiveLines(TrainData)) ); i < (MINIBATCH_SIZE+offset) * double(d->positiveLines(TrainData));i++){
        //retVal += pf[omp_get_thread_num()]->acomputeFit(d->getPositiveLine(i,TrainData));
        retVal += privateRnn->acomputeFit(d->getPositiveLine(i,TrainData));
        total += adouble(d->getPositiveLine(i,TrainData).size())-1.0;
    }
    for(int i = offset*double(d->negativeLines()); i < (MINIBATCH_SIZE+offset) * double(d->negativeLines());i++){
        //pf[omp_get_thread_num()]->setParameters(x);
        //retVal += (d->getNegativeLine(i,TrainData).size() - pf[omp_get_thread_num()]->acomputeFit(d->getNegativeLine(i,TrainData)));
        retVal += (d->getNegativeLine(i,TrainData).size() - privateRnn->acomputeFit(d->getNegativeLine(i,TrainData)));
        total += adouble(d->getNegativeLine(i,TrainData).size())-1.0;
    }
    //std::cout << retVal/total << " " << retVal << "/" << total << " ";
    return  retVal/(total);
}

Data adam::getResult(){
    return  newW;
}
