#include "classifier.h"
#include <cstring>
#include <algorithm>

Classifier::Classifier(nikolas::rnn *RnInst, Dataset *trainD){
    Rnn = RnInst;
    d = trainD;
    results = new bool [trainD->positiveLines()+trainD->negativeLines()];
    logProbs = new float [trainD->positiveLines()+trainD->negativeLines()];
    logProbsSorted = new float [trainD->positiveLines()+trainD->negativeLines()];
}

Classifier::~Classifier(){
    delete [] results;
    delete [] logProbs;
    delete [] logProbsSorted;
}

void Classifier::setWeights(int pos, float val){
    if(pos == 0) ClassWeights[0][0] = val;
    else if (pos == 1) ClassWeights[0][1] = val;
    else if (pos == 2) ClassWeights[1][0] = val;
    else if (pos == 3) ClassWeights[1][1] = val;
}

void Classifier::computeProbs(){
    for(unsigned i = 0; i < d->positiveLines();i++)
        logProbs[i] = Rnn->logProb(d->getPositiveLine(i));
    for(unsigned i = 0; i < d->negativeLines();i++)
        logProbs[i+d->positiveLines()] = Rnn->logProb(d->getNegativeLine(i));
    memcpy(logProbsSorted,logProbs,sizeof (float) * (d->positiveLines()+d->negativeLines()));
    std::sort(logProbsSorted, logProbsSorted + d->positiveLines()+d->negativeLines() );

}

void Classifier::computeThreshold(){
    computeProbs();
    float minThresh = logProbsSorted[0];
    float minError = 1e+10;
    float tempError;
    unsigned FP = 0, FN = 0;
    for(unsigned i = 0; i < d->positiveLines()+d->negativeLines()-1;i++){
        thresh = (logProbsSorted[i] + logProbsSorted[i+1])/2;
        tempError=0.0;
        FP = 0;
        FN = 0;
        for(unsigned j=0;j < d->positiveLines()+d->negativeLines(); j++){
            if(j < d->positiveLines()){
                if(logProbs[j] >= thresh){
                    tempError-=ClassWeights[0][0];
                }
                else {
                    tempError += ClassWeights[0][1];
                    FP++;
                }
            }
            else{
                if(logProbs[j] < thresh){
                    tempError -= ClassWeights[1][1];
                }
                else {
                    tempError += ClassWeights[1][0];
                    FN++;
                }
            }
        }
        //printf("%f %f %f %f\n",minError,tempError,minThresh,thresh);
        if(tempError<minError){
            minError = tempError;
            minThresh = thresh;
        }
        else if (tempError==minError && FP<FN) {
            minError = tempError;
            minThresh = thresh;
        }
    }
    thresh = minThresh;
}

void Classifier::printResults(){
    TestResults[0] = TestResults[1] = 0;
    int TrainRes[2][2]={{0,0},{0,0}};
    int TestRes[2][2]={{0,0},{0,0}};
    for(unsigned i = 0; i < d->positiveLines()+d->negativeLines();i++){
        if(i<d->positiveLines()){
            if(logProbs[i]>=thresh) TrainRes[0][0]++;
            else TrainRes[0][1]++;
        }
        else{
            if(logProbs[i]>=thresh) TrainRes[1][0]++;
            else TrainRes[1][1]++;
        }
    }
    for(unsigned i = 0; i < d->positiveLines(TestData)+d->negativeLines(TestData);i++){
        if(i<d->positiveLines(TestData)){
            if(Rnn->logProb(d->getPositiveLine(i,TestData)) >=thresh) TestRes[0][0]++;
            else TestRes[0][1]++;
        }
        else{
            if(Rnn->logProb(d->getNegativeLine(i-d->positiveLines(TestData),TestData))>=thresh) TestRes[1][0]++;
            else TestRes[1][1]++;
        }
    }
    TestResults[0] = (1.0*TestRes[0][0]) / d->positiveLines(TestData);
    TestResults[1] = (1.0*TestRes[1][1]) / d->negativeLines(TestData);
    //printf("%f\n",thresh);
    puts(" >> Classification Results\n   Train             Test      ");
    puts("+-------------+  +-------------+");
    printf("| %4d | %4d |  | %4d | %4d |\n",TrainRes[0][0],TrainRes[0][1],TestRes[0][0],TestRes[0][1]);
    puts("+-------------+  +-------------+");
    printf("| %4d | %4d |  | %4d | %4d |\n",TrainRes[1][0],TrainRes[1][1],TestRes[1][0],TestRes[1][1]);
    puts("+-------------+  +-------------+");
    std::cout << "Test Accuracy : " << (1.0*TestRes[0][0]+1.0*TestRes[1][1])/(1.0*TestRes[0][0]+1.0*TestRes[0][1]+1.0*TestRes[1][0]+1.0*TestRes[1][1]) << std::endl;
}

void Classifier::setRnn(nikolas::rnn *r){
    Rnn = r;
}

void Classifier::writeResults(vector<float> &TP, vector<float> &TN ){
    TP.push_back(TestResults[0]);
    TN.push_back(TestResults[1]);
}
