#include "dataset.h"
# define MAXLINESIZE 1024
# define TESTSIZE 0.2 //percentage of data used for testing
#include <algorithm>
#include <random>
#include <math.h>

Dataset::Dataset(char *filename) : maxLen(0)
{
    //format*******
    //M (number of terminal symbols)
    //P (number of positive symbols)
    //N (number of negative symbols)
    //symbol1
    //symbol2
    //.......
    //symbolM
    //example1
    //example2
    //.....
    //exampleM
    FILE *fp=fopen(filename,"r");
    if(!fp) exit(fprintf(stderr,"File %s not ready for open \n",filename));
    int M;
    int P;
    int N;
    char line[MAXLINESIZE];
    fscanf(fp,"%d",&M);
    fscanf(fp,"%d",&P);
    fscanf(fp,"%d",&N);
	while(vocabulary.size()!=M)
	{
        fgets(line,MAXLINESIZE-1,fp);
	line[strlen(line)-1]=0;
	if(strlen(line)==0) continue;
        //for now we read just the first character...
        vocabulary.insert(line[0]);
	}
    buildDictionary();
    int linecount=1;
    while(!feof(fp))
    {
        fgets(line,MAXLINESIZE-1,fp);
        line[strlen(line)-1]=0;
        LINE d;
        d.resize(strlen(line));
        for(unsigned i=0;i<strlen(line);i++)
        {
            d[i]=symbolCode( line[i]);
        }
        maxLen = (d.size() > maxLen)?d.size():maxLen;
        if(linecount<=P) posData.push_back(d);
        else  negData.push_back(d);
        linecount++;
    }
    fclose(fp);
	
    //shuffle data for random sampling
    //*
    auto rng = std::default_random_engine {};
    std::shuffle(std::begin(posData), std::end(posData), rng);
    std::shuffle(std::begin(negData), std::end(negData), rng);//*/
    //copy data
    for(unsigned i = 0; i < (TESTSIZE*P);i++) testPosData.push_back(posData[i]);
    for(unsigned i = 0; i < (TESTSIZE*N);i++) testNegData.push_back(negData[i]);
    //erase
    posData.erase(posData.end()- (unsigned int) TESTSIZE*P,posData.end());
    negData.erase(negData.end()- (unsigned int) TESTSIZE*N,negData.end());
    posData.resize((unsigned) ((1.0 - (float)TESTSIZE)*P) );
    negData.resize((unsigned) ((1.0 - (float)TESTSIZE)*N) );
    if(Vars.verbose){
    puts("+-------------------------------------------+\n| TrainPos | TrainNeg |  TestPos |  TestNeg |");
    puts("+-------------------------------------------+");
    printf("| %8d | %8d | %8d | %8d |\n",posData.size(),negData.size(),testPosData.size(),testNegData.size());
    puts("+-------------------------------------------+");
    }
}

int     Dataset::positiveLines(int DataSet) const
{
    if(DataSet==TrainData) return posData.size();
    else if (DataSet==TestData) return testPosData.size();
    else return -1;
}

unsigned Dataset::getMaxLen() const{
    return maxLen;
}

int     Dataset::negativeLines(int DataSet) const
{
    if(DataSet==TrainData) return negData.size();
    else if (DataSet==TestData) return testNegData.size();
    else return -1;
}

const LINE    &Dataset::getPositiveLine(int pos,int DataSet) const
{
    if(DataSet==TrainData) return posData[pos];
    else return testPosData[pos];
}

const LINE    &Dataset::getNegativeLine(int pos,int DataSet) const
{
    if(DataSet==TrainData) return negData[pos];
    else return testNegData[pos];
}


int     Dataset::vocabularySize()
{
    return vocabulary.size();
}


void    Dataset::addSymbolsToVocabulary(LINE &input)
{

}

unsigned int     Dataset::symbolCode(SYMBOL c)
{
    return symbolToInt[c];
}


vector<unsigned> Dataset::lineToCode(LINE &l)
{
    vector<unsigned> x;
    for(unsigned i=0;i<l.size();i++)
        x.push_back(symbolCode(l[i]));
    return x;
}

void    Dataset::buildDictionary()
{
    intToSymbol.clear();
    symbolToInt.clear();
    unsigned i = 0;
    for (auto &c : vocabulary)
    {
         intToSymbol.push_back(c);
         symbolToInt.insert(std::pair<char,unsigned>(c, i));
         ++i;
     }
}



