#include "genetic.h"
# include <tolmin.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <cmath>
#include <omp.h>
# include <pso.h>
#include "classifier.h"
#include "adam.h"
#include "a_gsl_bfgs.h"


double mrand()
{
    return rand()*1.0/RAND_MAX;
}


double dmax(double a, double b)
{
    return a>b?a:b;
}

double randomDouble()
{
    return 2*(mrand()-0.5);
}

Genetic::Genetic(nikolas::rnn *r,Dataset *t,int count,int numthreads)
: generation(0), numThreads(numthreads)
{
    //rnn=r;
    train=t;
    stack = new adept::Stack();
    chromosome.resize(count);
    fitnessArray.resize(count);
    children.resize(count);
    
    setSeletionRate(0.10);
    setMutationRate(0.05);
    rnn2 = r;
    int size=rnn2->parameterSize();

    rnn3 = new nikolas::rnn * [numThreads];

    for(unsigned i = 0; i < numThreads; i++){
        rnn3[i] = new nikolas::rnn(train->vocabularySize(),r->getHiddenSize());
    }

    for(int i=0;i<count;i++)
    {
        chromosome[i].resize(size);
        children[i].resize(size);
        for(int j=0;j<size;j++)
        {
            chromosome[i][j]=(randomDouble());
        }
        //fitnessArray[i]=fast_fitness(chromosome[i]); // <- redundant

    }
    c = new Classifier(rnn3[0],train);
    tempx.resize(size);

    results.resize(5,std::vector<double>());
    ClassResults.resize(2,std::vector<float>());

}

Genetic::~Genetic(){
    string out(string("Results")+Vars.filename.toStdString()+string(".png"));
    string out1(string("ClassResults")+Vars.filename.toStdString()+string(".png"));
    delete rnn2;
    for(unsigned i = 0; i < numThreads;i++) delete rnn3[i];
    stack->pause_recording();
    stack->clear_gradients();
    delete stack;
}

int     Genetic::getChromosomes() const
{
    return chromosome.size();
}

double  Genetic::getSelectionRate() const
{
    return selection_rate;
}

void    Genetic::setSeletionRate(double r)
{
    if(r>=0 && r<=1)
        selection_rate=r;
}

double  Genetic::getMutationRate() const
{
    return mutation_rate;
}

void    Genetic::setMutationRate(double r)
{
    if(r>=0 && r<=1)
        mutation_rate=r;
}

void    Genetic::nextGeneration()
{

    if(generation) mutation();
    calcFitnessArray();

    selection();

    crossover();
    //crossover <- went to end of function body
    ++generation;
    //local search
    delete stack;
    if(generation % (Vars.LocalSearch)==0 && generation)
	{
        int count=Vars.LocalSearchTries;
        int localThreads = numThreads;
#pragma omp parallel for num_threads(localThreads)
        for(int i=0;i<count;i++)
        {
            int randPos=rand() % chromosome.size();
            localSearch(randPos);
        }
        selection();
        chromosome.back() = chromosome[0];
        localSearch(chromosome.size()-1);
        selection();
    }
    stack = new adept::Stack();
    rnn3[0]->setParameters(chromosome[0]);
    c->setRnn(rnn3[0]);
    c->computeThreshold();
    if(Vars.verbose) c->printResults();
    c->writeResults(ClassResults[0],ClassResults[1]);

    results[0].push_back(fitnessArray[0]);
    results[1].push_back(fitnessArray[0.15*fitnessArray.size()]);
    results[2].push_back(fitnessArray[0.3*fitnessArray.size()]);
    results[3].push_back(fitnessArray[0.45*fitnessArray.size()]);
    results[4].push_back(fitnessArray[0.6*fitnessArray.size()]);

}

int     Genetic::getGeneration() const
{
    return generation;
}

void    Genetic::getBest(Data &x,double &y)
{
    x=chromosome[0];
    y=fitnessArray[0];
}

void    Genetic::setBest(Data &x,double &y)
{
    chromosome[0]=x;
    fitnessArray[0]=y;
}

void    Genetic::granal(Data &x,Data &g)
{
    unsigned j,i;
    double v=fast_fitness(x,omp_get_thread_num());
//#pragma omp parallel for private(i) shared(g) firstprivate(v,x) num_threads(numThreads)
    for(i=0;i<x.size();i++)
    {
        double eps=pow(1e-18,1.0/3.0)*dmax(1.0,fabs(x[i]));
        x[i]+=eps;
        double v1=fast_fitness(x,omp_get_thread_num());
        g[i]=(v1-v)/(1.0 * eps);
        x[i]-=eps;
    }
}

double Genetic::ProbAccuracy(Data &x, int DataSet, unsigned net){

    int OriginSet;
    unsigned posLines;
    if(DataSet==TrainData){
        posLines=train->positiveLines(TrainData);
        OriginSet = TrainData;
    }
    else {
        posLines=train->positiveLines(TestData);
        OriginSet = TestData;
    }
    int total_line_size=0;
    double total_loss=0;
    unsigned int i;

    rnn3[net]->setParameters(x);
    //Testrnn2->setParameters(x);
    LINE temp;
    vector<unsigned> lineat;
    for(i=0;i<posLines;i++)
    {
      temp = train->getPositiveLine(i,OriginSet);
      //lineat=train->lineToCode(temp);
      total_loss+=rnn3[net]->computeError(temp);
      total_line_size+=temp.size()-1;
     }
    if(total_line_size==0) total_line_size++;

    double p1=total_loss*1.0/total_line_size;
    return p1;
}

bool tolminflag=false;

double Genetic::NaiveAccuracy(Data &x, int DataSet, unsigned net){
    int OriginSet;
    unsigned posLines;
    if(DataSet==TrainData){
        posLines=train->positiveLines(TrainData);
        OriginSet = TrainData;
    }
    else {
        posLines=train->positiveLines(TestData);
        OriginSet = TestData;
    }
    int total_line_size=0;
    double total_loss=0;
    unsigned int i;

    //nikolas::rnn *Testrnn2 = new nikolas::rnn(train->vocabularySize(),10);
    //Testrnn2->setParameters(x);
    rnn3[net]->setParameters(x);
    LINE temp;
    vector<unsigned> lineat;
    for(i=0;i<posLines;i++)
    {
      temp = train->getPositiveLine(i,OriginSet);
      //lineat=train->lineToCode(temp);
      total_loss+=rnn3[net]->computeNaive(temp);
      total_line_size+=temp.size()-1;
     }

    //delete  Testrnn2;
    if(total_line_size == 0) total_line_size++;
    double p1=total_loss*1.0/total_line_size;
    return p1;

}

double Genetic::NegativeNaive(Data &x, int DataSet, unsigned net){

    int OriginSet;
    unsigned negLines;
    if(DataSet==TestData){
        negLines=train->negativeLines(TestData);
        OriginSet = TestData;
    }
    else {
        negLines=train->negativeLines(TrainData);
        OriginSet = TrainData;
    }
    int total_line_size=0;
    double total_loss=0;
    unsigned int i;

    //nikolas::rnn *Testrnn2 = new nikolas::rnn(train->vocabularySize(),10);
    //Testrnn2->setParameters(x);
    rnn3[net]->setParameters(x);
    LINE temp;
    vector<unsigned> lineat;
    for(i=0;i<negLines;i++)
    {
      temp = train->getNegativeLine(i,OriginSet);
      //lineat=train->lineToCode(temp);
      total_loss+=rnn3[net]->computeNaive(temp);
      total_line_size+=temp.size()-1;
     }

    //delete  Testrnn2;
    if(total_line_size == 0) total_line_size++;
    double p1=total_loss*1.0/total_line_size;
    return p1;

}
double Genetic::NegativeProb(Data &x, int DataSet, unsigned net){
    int OriginSet;
    unsigned negLines;
    if(DataSet==TestData){
        negLines=train->negativeLines(TestData);
        OriginSet = TestData;
    }
    else {
        negLines=train->negativeLines(TrainData);
        OriginSet = TrainData;
    }
    int total_line_size=0;
    double total_loss=0;
    unsigned int i;

    //nikolas::rnn *Testrnn2 = new nikolas::rnn(train->vocabularySize(),10);
    //Testrnn2->setParameters(x);
    rnn3[net]->setParameters(x);
    LINE temp;
    vector<unsigned> lineat;
    for(i=0;i<negLines;i++)
    {
      temp = train->getNegativeLine(i,OriginSet);
      //lineat=train->lineToCode(temp);
      total_loss+=rnn3[net]->computeError(temp);
      total_line_size+=temp.size()-1;
     }

    //delete  Testrnn2;
    if(total_line_size == 0) total_line_size++;
    double p1=total_loss*1.0/total_line_size;
    return p1;
}

double  Genetic::fitness(Data &x, unsigned net)
{
    return 0.25*(ProbAccuracy(x,TrainData,net) + NaiveAccuracy(x,TrainData,net) + 2.0 - NegativeProb(x,TrainData,net)-NegativeNaive(x,TrainData,net));

}

double Genetic::fast_fitness(Data &x, unsigned net){

    double error = 0.0;
    int total_line_size=0;
    //LINE temp;
    rnn3[net]->setParameters(x);
    //vector<unsigned> lineat;
    int sizeL;
    for(unsigned i = 0; i < train->positiveLines();i++){
        //temp = train->getPositiveLine(i,TrainData);
        //lineat=train->lineToCode(temp);
        //error+= rnn3[net]->computeNaive(temp);
        error+= rnn3[net]->computeFit(train->getPositiveLine(i,TrainData));
        total_line_size+=(train->getPositiveLine(i,TrainData).size()-1);
    }
    for(unsigned i = 0; i < train->negativeLines();i++){
        //temp = train->getNegativeLine(i,TrainData);
        //lineat=train->lineToCode(temp);
        sizeL = train->getNegativeLine(i,TrainData).size();
        error+= sizeL - rnn3[net]->computeFit(train->getNegativeLine(i,TrainData));
        total_line_size+=sizeL-1;
    }
    return error/total_line_size;
}

void    Genetic::localSearch(int pos)
{
    Data tempx(rnn3[omp_get_thread_num()]->parameterSize());
    int size=chromosome[0].size();
    int count=chromosome.size();
    for(int iters=1;iters<=100;iters++)
    {
       int gpos=rand() % count;
       if(Vars.LocalMethod==Uniform){

       for(int j=0;j<size;j++)
       {
           double alfa=-0.5+2.0*mrand();
           double v=alfa*chromosome[pos][j]+(1.0-alfa)*chromosome[gpos][j];
            tempx[j]=v;
       }
       double f=fast_fitness(tempx,omp_get_thread_num());
       if(fabs(f)<fabs(fitnessArray[pos]))
       {
           if(Vars.verbose) printf("NEW MIN[%4d]=%10.4lg [THREAD : %d]\n",pos,f,omp_get_thread_num());
           for(int j=0;j<size;j++) chromosome[pos][j]=tempx[j];
           fitnessArray[pos]=f;
        }
       }
       if(Vars.LocalMethod==Crossover){
       //*
       int cutpoint=rand() % size;
       for(int j=0;j<cutpoint;j++)    tempx[j]=chromosome[pos][j];
       for(int j=cutpoint;j<size;j++) tempx[j]=chromosome[gpos][j];
       double f=fast_fitness(tempx,omp_get_thread_num());
       if(fabs(f)<fabs(fitnessArray[pos]))
       {
           if(Vars.verbose) printf("NEW MIN[%4d]=%10.4lg [THREAD : %d]\n",pos,f,omp_get_thread_num());
           for(int j=0;j<size;j++) chromosome[pos][j]=tempx[j];
           fitnessArray[pos]=f;
        }
        else
        {
          for(int j=0;j<cutpoint;j++) tempx[j]=chromosome[gpos][j];
          for(int j=cutpoint;j<size;j++) tempx[j]=chromosome[pos][j];
          double f=fast_fitness(tempx,omp_get_thread_num());
          if(fabs(f)<fabs(fitnessArray[pos]))
          {
             if(Vars.verbose) printf("NEW MIN[%4d]=%10.4lg [THREAD : %d]\n",pos,f,omp_get_thread_num());
             for(int j=0;j<size;j++) chromosome[pos][j]=tempx[j];
             fitnessArray[pos]=f;
            }
         }//*/
        }
      }
     if(Vars.LocalMethod==AdaGrad)
     {
         Data g,tempx1;
         g.resize(chromosome[0].size());

         double rate=0.001;//rnn2->learnRate();
         double f;
         tempx1=chromosome[pos];
         const int maxiters=20;
         for(int iter=1;iter<=maxiters;iter++)
         {
            granal(tempx1,g);
            double sum=0.0;
            for(unsigned i=0;i<g.size();i++)
            {
                double xx=tempx1[i];
                 tempx1[i]=tempx1[i]-rate * g[i];
                 sum+=pow(xx-tempx1[i],2.0);
            }

            f=fast_fitness(tempx1,omp_get_thread_num());
            if(Vars.verbose) printf("ITER:%d OLD FITNESS: %20.8lf NEW FITNESS: %20.8lf [THREAD : %d]\n",iter,fitnessArray[pos],f,omp_get_thread_num());
            if(sum<1e-10) break;
#pragma omp critical
{
            if(f<fitnessArray[pos])
            {
                fitnessArray[pos]=f;
                chromosome[pos]=tempx1;
            }
}
         }
        }
        if(Vars.LocalMethod==Tolmin)
        {
            Data tempx1=chromosome[pos];
            MinInfo Info;
            Info.p=this;
            Info.iters=Vars.bfgsIters;
            Info.pf=rnn3;
            Info.d=train;
            double f;
            f=tolmin(tempx1,Info);
#pragma omp critical
{
            if(f<fitnessArray[pos])
            {
                chromosome[pos]=tempx1;
                fitnessArray[pos]=f;
            }
}
        }
        if(Vars.LocalMethod==Adam)
        {
            //std::cout << fitnessArray[pos] << " Before Adam" << std::endl;
            adam a(this,rnn3,chromosome[pos],Vars.adamIterations,train,Vars.numberOfThreads);
            a.run();
#pragma omp critical
{
            chromosome[pos] = a.getResult();
            fitnessArray[pos] = fast_fitness(chromosome[pos]);
            std::cout << fitnessArray[pos] << " After Adam THREAD: " << omp_get_thread_num() << std::endl;//*/
}
            //delete a;

        }

}


void	Genetic::bfgs(int pos)
{
	tolminflag=true;
            tempx=chromosome[pos];
            MinInfo Info;
            Info.p=this;
            Info.iters=Vars.bfgsIters;
            double f=tolmin(tempx,Info);
            if(f<fitnessArray[pos])
            {
                chromosome[pos]=tempx;
                fitnessArray[pos]=f;
            }
	tolminflag=false;
}

void    Genetic::calcFitnessArray()
{
#pragma omp parallel for num_threads(numThreads)
    for(int i = 0; i < chromosome.size();i++)
    {
        fitnessArray[i]=fast_fitness(chromosome[i],omp_get_thread_num());
    }
   }

void    Genetic::selection()
{
    double t;
    for(unsigned i=0;i<chromosome.size();i++)
    {
        for(unsigned j=0;j<chromosome.size()-1;j++)
        {
            if(fitnessArray[j+1]<fitnessArray[j])
            {
                t=fitnessArray[j];
                fitnessArray[j]=fitnessArray[j+1];
                fitnessArray[j+1]=t;
                tempx=chromosome[j];
                chromosome[j]=chromosome[j+1];
                chromosome[j+1]=tempx;
            }
        }
    }
}

int     Genetic::tournament()
{
    const int tournament_size=8;
    double minFitness=1e+100;
    int minIndex=-1;
    for(int i=1;i<=tournament_size;i++)
    {
        int randomPos=rand() % chromosome.size();
        if(fitnessArray[randomPos]<minFitness)
        {
            minFitness=fitnessArray[randomPos];
            minIndex=randomPos;
        }
    }
    return minIndex;
}

void    Genetic::crossover()
{
    int children_size=(int)((1.0-selection_rate)*chromosome.size());
    int children_count=0;
    while(true)
    {
        int parent1=tournament();
        int parent2=tournament();
        int size=chromosome[0].size();
        for(int i=0;i<size;i++)
        {
            double x1,x2,g1,g2;
            x1=chromosome[parent1][i];
            x2=chromosome[parent2][i];
            double alfa=-0.5+2.0*mrand();
            g1=alfa*x1+(1.0-alfa)*x2;
            g2=alfa*x2+(1.0-alfa)*x1;
            children[children_count][i]=g1;
            children[children_count+1][i]=g2;
        }
        children_count+=2;
        if(children_count>=children_size) break;
    }
    for(int i=0;i<children_count;i++)
    {
        chromosome[chromosome.size()-i-1]=children[i];
    }
}

void    Genetic::mutation()
{
    int size=chromosome[0].size();
    for(unsigned i=1;i<chromosome.size();i++)
    {
        for(int j=0;j<size;j++)
        {
            double r=mrand();
            if(r<mutation_rate)
            {
                chromosome[i][j]=chromosome[i][j]*(1.0 +0.05* randomDouble());
            }
        }
    }
}


void Genetic::printResults()
{
	c->printResults();
}

inline void process_mem_usage(double& vm_usage, double& resident_set)
{
    vm_usage     = 0.0;
    resident_set = 0.0;
    // the two fields we want
    unsigned long vsize;
    long rss;
    {
        std::string ignore;
        std::ifstream ifs("/proc/self/stat", std::ios_base::in);
        ifs >> ignore >> ignore >> ignore >> ignore >> ignore >> ignore >> ignore >> ignore >> ignore >> ignore
            >> ignore >> ignore >> ignore >> ignore >> ignore >> ignore >> ignore >> ignore >> ignore >> ignore
            >> ignore >> ignore >> vsize >> rss;
        }
        long page_size_kb = sysconf(_SC_PAGE_SIZE) / 1024; // in case x86-64 is configured to use 2MB pages
        vm_usage = vsize / 1024.0;
        resident_set = rss * page_size_kb;
}

inline float getMem(){
    double vm, rss;
    process_mem_usage(vm, rss);
    return rss/8192.0;
}

inline void printMemory(){
    double vm, rss;
    process_mem_usage(vm, rss);
    std::cout << "VM: " << vm/1024.0 << " Mbits; RSS: " << rss/1024.0 << " Mbits" << std::endl;
} //*/
