#ifndef IMPROVED_RNN_H
#define IMPROVED_RNN_H
#include <adept.h>
#include <vector>
#include <dataset.h>

using adept::afloat;
using adept::adouble;

namespace nikolas {
class rnn{
    public :
        // Just memory allocation //
        rnn(unsigned lexicSize, unsigned hiddenSize);
        ~rnn();
        // Interface for the GA //
        double computeError(vector<unsigned> &in);
        double computeNaive(vector<unsigned> &in);
        double computeFit(const vector<unsigned> &in);
        adouble acomputeFit(const vector<unsigned> &in);
        void setParameters(std::vector<double> &x);
        void setParameters(std::vector<adept::adouble> &x);
        unsigned parameterSize() const;
        unsigned getHiddenSize() const;
        unsigned getLexicSize() const;
        std::vector<unsigned> getState();
        float logProb(const vector<unsigned> &in);

    private :
        // Size of RNN //
        unsigned hiddenSize;
        unsigned lexicSize;
        // Internal weights //
        float **Wxh;
        float **Whh;
        float **Why;
        float *by;
        float *bh;
        // Hidden state vector //
        float *Hprev;
        // Temp Vars //
        float *xs;
        float *ys;
        float *tempHs;
        // ADEPT VARS
        // Internal weights //
        adouble **aWxh;
        adouble **aWhh;
        adouble **aWhy;
        adouble *aby;
        adouble *abh;
        // Hidden state vector //
        adouble *aHprev;
        // Temp Vars //
        adouble *axs;
        adouble *ays;
        adouble *atempHs;

        void updateaHs();
        void computeaYs();

        // Compatibility vars //
        std::vector<unsigned> output;
        // Private computations //
        void updateHs();
        void computeYs();
        unsigned pointCounter;

    };
}

#endif // IMPROVED_RNN_H
