# include <iostream>
# include <dataset.h>
# include <genetic.h>
# include <stdlib.h>
# include <optionsparser.h>
# include <adept.h>
# include <string>
# include <QString>
# include <QJsonObject>
# include <QStringList>
# include <QDebug>
# include <QCoreApplication>
# define VERSION "1.0.1"
using namespace std;


//adept::Stack stack;

GlobalVars Vars;
QJsonObject params;

void printHelp()
{
    std::cout << " This program trains an RNN Classifier" << std::endl;
    std::cout << " Options [=default]:" << std::endl;
    std::cout << "\t 1) General options :" << std::endl;
    std::cout << "\t\t --threads=n [=4] Specify the number of threads" << std::endl;
    std::cout << "\t\t --seed=s [=1] Specify the seed for random initialization" << std::endl;
    std::cout << "\t\t --filename=dataset.txt [=data.txt] Specify the dataset filename" << std::endl;
    std::cout << "\t\t --help Print the help message" << std::endl;
    std::cout << "\t\t --version Print the version of the software" << std::endl;
    std::cout << "\t\t --verbose=[0|1[=1]] Choose if everything will be printed in the terminal" << std::endl;
    std::cout << "\t 2) Genetic Algorithm options :" << std::endl;
    std::cout << "\t\t --chromosomes=n [=500] Specify the number of chromosomes" << std::endl;
    std::cout << "\t\t --generations=n [=300] Specify the number of generations" << std::endl;
    std::cout << "\t\t --selection_rate=p" << std::endl;
    std::cout << "\t\t --mutation_rate=p" << std::endl;
    std::cout << "\t 3) Local optimization options :" << std::endl;
    std::cout << "\t\t --localsearch_method=[UNIFORM|CROSSOVER|ADAGRAD|TOLMIN|ADAM[=TOLMIN]]" << std::endl;
    std::cout << "\t\t --localsearch_iters=n [=20] Every n generations local search is activated" << std::endl;
    std::cout << "\t\t --localsearch_chromosomes=n [=10] Specify the number of chromosomes subject to local optimization" << std::endl;
    std::cout << "\t\t --bfgs_iters=n [=100] Specify the maximum iterations of BFGS or TOLMIN" << std::endl;
    std::cout << "\t\t --adam_iters=n [=100] Specify the maximum iterations of ADAM optimizer" << std::endl;
    std::cout << "\t\t --pso_iters=n [=100] Specify the maximum iterations of PSO" << std::endl;
    std::cout << "\t\t --pso_particles=n [=100] Specify the number of particles for PSO" << std::endl;
    std::cout << "\t 4) RNN classifier options :" << std::endl;
    std::cout << "\t\t --number_of_weights=n [=10] Specify the hidden size of the RNN" << std::endl;

}


void    printVersion()
{
    qDebug()<<"Genetic Recurrent Neural Network\nVersion: "<<VERSION<<endl;
}


void parseCmdLine(QStringList args)
{
    QString lastParam="";
    QStringList localMethod;

    localMethod<<"UNIFORM"<<"CROSSOVER"<<"ADAGRAD"<<"TOLMIN"<<"ADAM";

    for(int i=1;i<args.size();i++)
    {
        QStringList sp=args[i].split("=");
        if(sp[0]=="--version")
          std::cout << "GeneticRNN version : " << VERSION << std::endl;
        else
        if(sp[0]=="--verbose")
          Vars.verbose = sp[1].toInt();
        else
        if(sp[0]=="--filename")
          Vars.filename = sp[1];
        else
        if(sp[0]=="--seed")
            Vars.randomSeed=sp[1].toInt();
        else
        if(sp[0]=="--chromosomes")
            Vars.populous=(size_t) sp[1].toInt();
        else
        if(sp[0]=="--generations")
            Vars.generations=(size_t)sp[1].toInt();
        else
        if(sp[0]=="--selection_rate")
            Vars.selectionRate=sp[1].toDouble();
        else
        if(sp[0]=="--mutation_rate")
            Vars.mutationRate=sp[1].toDouble();
        else
        if(sp[0]=="--localsearch_iters")
            Vars.LocalSearch=sp[1].toInt();
        else
        if(sp[0]=="--pso_iters")
            Vars.psoGenerations=sp[1].toInt();
        else
        if(sp[0]=="--pso_particles")
            Vars.psoParticles=sp[1].toInt();
        else
        if(sp[0]=="--localsearch_chromosomes")
            Vars.LocalSearchTries=sp[1].toInt();
        else
        if(sp[0]=="--localsearch_method")
          Vars.LocalMethod=(localMethod.indexOf(sp[1])==-1)?Adam:
                                 localMethod.indexOf(sp[1]);
        else
        if(sp[0]=="--number_of_threads")
            Vars.numberOfThreads=sp[1].toInt();
        else
        if(sp[0]=="--number_of_weights")
            Vars.numberOfWeights=sp[1].toInt();
        else
        if(sp[0]=="--threads")
             Vars.numberOfThreads=sp[1].toInt();
        else
        if(sp[0]=="--bfgs_iters")
            Vars.bfgsIters=sp[1].toInt();
        else
        if(sp[0]=="--adam_iters")
            Vars.adamIterations=sp[1].toInt();
        else
        if(sp[0]=="--help")
        {
            printHelp();
            exit(EXIT_SUCCESS);
        }
        else
        if(sp[0]=="--version")
        {
            printVersion();
            exit(EXIT_SUCCESS);
        }
        else
        if(sp[0]=="--param" || sp[0]=="-p")
            lastParam=sp[1];
        else
        if(sp[0]=="--value" "" || sp[0]=="-v")
        {
            //thre cases
            //int
            //double
            //string
            //all double numbers should have .
            bool assigned=false;
            if(sp[1].contains("."))
            {
                bool b;
                qreal f=sp[1].toDouble(&b);
                if(b) {
                    assigned=true;
                    params[lastParam]=f;
                }
            }

            if(!sp[1].contains("."))
            {
                bool b1;
                qint64 ix=sp[1].toInt(&b1);
                if(b1){assigned=true; params[lastParam]=ix;}
            }
            if(!assigned)
            params[lastParam]=sp[1];
        }
        else
        {
            qDebug()<<"Wrong parameter "<<sp[0]<<endl;
            exit(printf("Wrong arguments\n"));
        }
    }
}

int main(int argc, char **argv)
{
    //adept::Stack stack;
    QCoreApplication app(argc,argv);
    parseCmdLine(app.arguments());
    adept::set_max_blas_threads(Vars.numberOfThreads);
    //stack.set_max_jacobian_threads(Vars.numberOfThreads);
    srand(Vars.randomSeed);
    char *fname = (char *)Vars.filename.toStdString().c_str();
    Dataset train(fname);
    adept::Stack *s = new adept::Stack();
    nikolas::rnn *Rnn = new nikolas::rnn(train.vocabularySize(),Vars.numberOfWeights);
    delete s;
    Genetic mgenetic(Rnn,&train,Vars.populous,Vars.numberOfThreads);
    mgenetic.setMutationRate(Vars.mutationRate);
    mgenetic.setSeletionRate(Vars.selectionRate);
    for(int g=1;g<=Vars.generations;g++)
    {
        mgenetic.nextGeneration();
        Data x;
        double y;
        x.resize(Rnn->parameterSize());
        mgenetic.getBest(x,y);
        if(Vars.verbose) printf("Generation=%03d Fitness=%10.4lf\n",g,y);
    }
    mgenetic.printResults();
    return 0;
}
