# include <pso.h>
# include <math.h>
# include <omp.h>
# include <optionsparser.h>
extern GlobalVars Vars;

static double dmin(double a,double b)
{	
	return a<b?a:b;
}


static double dmax(double a,double b)
{
	return a>b?a:b;
}

static double getNorm(Data x)
{
	double s=0.0;
	for(int i=0;i<x.size();i++) s+=x[i]*x[i];
	return sqrt(s);
}

static double getDistance(Data x1,Data x2)
{
	double s=0.0;
	for(int i=0;i<x1.size();i++) s+=(x1[i]-x2[i])*(x1[i]-x2[i]);
	return s;
}

static double maxElement(Data x)
{
	double m=-1e+100;
	for(int i=0;i<x.size();i++) if(x[i]>m) m=x[i];
	return m;
}


static double minElement(Data x)
{
	double m=1e+100;
	for(int i=0;i<x.size();i++) if(x[i]<m) m=x[i];
	return m;
}

double Pso::fitnessFunction(Data &x, unsigned net){
    double error = 0.0;
    int total_line_size=0;
    LINE temp;
    problem[net]->setParameters(x);
    vector<unsigned> lineat;
    int sizeL;

    for(unsigned i = 0; i < train->positiveLines();i++){
        //temp = train->getPositiveLine(i,TrainData);
        //lineat=train->lineToCode(temp);
        error+= problem[net]->computeFit(train->getPositiveLine(i,TrainData));
        total_line_size+=train->getPositiveLine(i,TrainData).size()-1;
    }

    for(unsigned i = 0; i < train->negativeLines();i++){
        //temp = train->getNegativeLine(i,TrainData);
        //lineat=train->lineToCode(temp);
        sizeL = train->getNegativeLine(i,TrainData).size();
        error+= sizeL - problem[net]->computeFit(train->getNegativeLine(i,TrainData));
        total_line_size+=sizeL-1;
    }
    return error/total_line_size;
}

double psorand()
{
    return rand() *1.0/RAND_MAX;
}

Pso::Pso(nikolas::rnn **p, Data &initialx, int count, Dataset *tr, unsigned Threads)
    :numThreads(Threads)
{
    train=tr;
	problem = p;
    int n=p[0]->parameterSize();
    xbest.resize(n);
	ybest   = 1e+100;
	generation = 0;
	genome.resize(count);
	fitness.resize(count);
	bestGenome.resize(count);
	bestFitness.resize(count);	
	oldBestFitness.resize(count);	
	vel.resize(count);
    lmargin.resize(n);
    rmargin.resize(n);
    center.resize(n);
    for(int i=0;i<n;i++)
    {
        lmargin[i]=-Vars.psoSphere * fabs(initialx[i]);
        rmargin[i]= Vars.psoSphere * fabs(initialx[i]);

    }
	for(int i=0;i<count;i++)
	{
        genome[i].resize(n);
        vel[i].resize(n);
        for(int j=0;j<vel[i].size();j++)
        {
            genome[i][j]=lmargin[j]+psorand()*(rmargin[j]-lmargin[j]);
            vel[i][j]=0.0;
        }
        bestGenome[i].resize(n);


		bestGenome[i]=genome[i];
        fitness[i]=fitnessFunction(genome[i]);
		bestFitness[i]=fitness[i];
		oldBestFitness[i]=bestFitness[i];
		if(fitness[i]<ybest)
		{
			ybest = fitness[i];
			xbest = genome[i];
			bestIndex = i;
		}
	}
    newGenerationFlag=false;
	x1=0.0;
	x2=0.0;
	oldbest = 1e+100;
	stopat=0.0;
	iters=2;
}


void    Pso::calcFitnessArray()
{
#pragma omp parallel for num_threads(numThreads)
    for(int i=0;i<genome.size();i++)
    {
        oldBestFitness[i]=bestFitness[i];

        fitness[i]=fitnessFunction(genome[i],omp_get_thread_num());

        if(fitness[i]<bestFitness[i])
        {
            bestFitness[i]=fitness[i];
            bestGenome[i]=genome[i];
        }
    }
    updateBest();
}


void	Pso::updateBest()
{
	for(int i=0;i<fitness.size();i++)
	{
		if(fitness[i]<ybest)
		{
			ybest = fitness[i];
			xbest = genome[i];
			bestIndex=i;
		}
	}

}

double	Pso::getInertia(int index)
{
	return 0.9-0.5*generation*1.0/generations;
}

void	Pso::boundVelocity(double &v,int index)
{
	const double gamma=0.5; //proposed by Ali
	double umax=gamma * (rmargin[index]-lmargin[index]);
	double umin=-umax;
	if(v<umin) v=umin;
	if(v>umax) v=umax;
}

void	Pso::updateCenter()
{
    for(int j=0;j<problem[0]->parameterSize();j++) center[j]=0.0;
	for(int i=0;i<genome.size();i++)
	{
        for(int j=0;j<problem[0]->parameterSize();j++)
		{
			center[j]+=genome[i][j];
		}
	}
    for(int j=0;j<problem[0]->parameterSize();j++)
	{
		center[j]/=genome.size();
		if(center[j]<lmargin[j]) center[j]=lmargin[j];
		if(center[j]>rmargin[j]) center[j]=rmargin[j];
	}
}

void	Pso::nextGeneration()
{
	++generation;
       newGenerationFlag=true;
	double psi1=1.0,psi2=1.0,psi3=1.0;
    Data xp;
    xp.resize(problem[0]->parameterSize());
    //double total_mean=0.0;
	double r1,r2,v1,v2,accel,accel_sign,trial_accel;
#pragma omp parallel for private(r1,r2,accel,v1,v2)
	for(int i=0;i<genome.size();i++)
	{

        for(int j=0;j<problem[0]->parameterSize();j++)
		{
            r1=psorand();
            r2=psorand();
            vel[i][j]=getInertia(i)*vel[i][j]+psi1*r1*(xbest[j]-genome[i][j])+
						psi2*r2*(bestGenome[i][j]-genome[i][j]);
            boundVelocity(vel[i][j],j);
            genome[i][j]=genome[i][j]+vel[i][j];

			if(genome[i][j]<lmargin[j]) 
			{	
				genome[i][j]=lmargin[j];
			}
			if(genome[i][j]>rmargin[j]) 
			{
				genome[i][j]=rmargin[j];
			}
		}
	}
}

void	Pso::calculateTerminationRule()
{
	double v=fabs(ybest+0.01);
	x1+=v;
	x2+=v*v;
	variance=x2/iters-(x1/iters)*(x1/iters);
	if(variance<0) variance=-variance;
	if(fabs(ybest-oldbest)>1e-4)
	{
		oldbest=ybest;
		stopat=variance/2.0;
	}
	if(stopat<1e-7) stopat=variance/2.0;
	iters++;
}


bool	Pso::terminationRule()
{
	if(generations<2)	return true;
	if(variance<stopat && generation>5) return true;
	return false;
}

void	Pso::Solve(int g,int iprint)
{
	generation=1;
	generations=g;
	for(int i=2;i<=g+1;i++)
	{
		if(terminationRule()) 
		{
			newGenerationFlag=true;
			break;
		}
		nextGeneration();
		calcFitnessArray();
        if(iprint) printf("ITERATION=%4d\tYBEST=%20.8lg\t",iters,ybest);
		calculateTerminationRule();
        if(iprint)
		printf("VARIANCE=%10.4lg\tSTOPAT=%10.4lg\n",variance,stopat);
	}
	newGenerationFlag=true;
}

int	Pso::getGeneration()
{
    return 0;
}

int	Pso::getGenomeCount()
{
	return genome.size();
}

int	Pso::getDimension()
{
    return problem[0]->parameterSize();
}

void	Pso::getBest(Data &x,double &y)
{
	x = xbest;
	y = ybest;
}

void	Pso::setBest(Data &x,double y)
{
	xbest=x;
	ybest=y;
}

Pso::~Pso()
{
}
